/**
 * The MIT License (MIT)
 *
 * Copyright (C) 2013 Sergi Mansilla
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the 'Software'), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

'use strict';

/**
 * Creates a virtually-rendered scrollable list.
 * @param {object} config
 * @constructor
 */
function VirtualList(config)
{
	var width = (config && config.w + 'px') || '100%';
	var height = (config && config.h + 'px') || '100%';
	var itemHeight = this.itemHeight = config.itemHeight;

	this.items = config.items;
	this.generatorFn = config.generatorFn;
	this.totalRows = config.totalRows || (config.items && config.items.length);

	var scroller = VirtualList.createScroller(itemHeight * this.totalRows);
	this.container = VirtualList.createContainer(width, height);
	this.container.appendChild(scroller);

	this.scroller = scroller;

	var screenItemsLen = Math.ceil(config.h / itemHeight);
	this.cachedItemsLen = screenItemsLen + 2;

	this.cachedRows = {};
	this.cacheSize = this.cachedItemsLen + 10;
	this.cacheIndex = [];

	var self = this;
	var lastRepaintY;
	var maxBuffer = itemHeight * 1;
	var lastScrolled = 0;

	function onScroll(e)
	{
		var scrollTop = e.target.scrollTop; // Triggers reflow
		if (!lastRepaintY || Math.abs(scrollTop - lastRepaintY) > maxBuffer)
		{
			var first = parseInt(scrollTop / itemHeight - 1);
			self._renderChunk(self.container, first < 0 ? 0 : first);
			lastRepaintY = scrollTop;
		}

		lastScrolled = Date.now();
		e.preventDefault && e.preventDefault();
	}

	this.container.addEventListener('scroll', onScroll);
}

VirtualList.prototype.init = function()
{
	this.scroller.style.height = (this.itemHeight * this.totalRows)+"px";
	this._renderChunk(this.container, 0);
}

VirtualList.prototype.cacheItem = function (i, item)
{
	this.cachedRows[i] = { "item": item, "index": i };
	this.cacheIndex.push(i);

	if (this.cacheIndex.length>this.cacheSize)
	{
		var index = this.cacheIndex[0];
		this.cacheIndex = this.cacheIndex.splice(1, this.cacheIndex.length - 1);
		var node = this.cachedRows[index].item;
		this.container.removeChild(node);
		delete this.cachedRows[index];
	}
}

function findIndex(ar,val)
{
	for (var i = 0;i < ar.length;i++)
	{
		if (ar[i] == val)
			return i;
	}
	return -1;
}

VirtualList.prototype.createRow = function (i)
{
	if (this.cachedRows[i] != null)
	{
		var index = findIndex(this.cacheIndex, i);
		this.cacheIndex.push(this.cacheIndex.splice(index, 1)[0]);
		var node = this.cachedRows[i].item;
        return { "cached": true, "item": node };
	}

	var item;
	if (this.generatorFn)
		item = this.generatorFn(i);
	else if (this.items)
	{
		if (typeof this.items[i] === 'string')
		{
			var itemText = document.createTextNode(this.items[i]);
			item = document.createElement('div');
			item.style.height = this.itemHeight + 'px';
			item.appendChild(itemText);
		} else
		{
			item = this.items[i];
		}
	}

	item.classList.add('vrow');
	item.style.position = 'absolute';
	item.style.top = (i * this.itemHeight) + 'px';

	this.cacheItem(i, item);

    return { "cached": false, "item": item };
};

/**
 * Renders a particular, consecutive chunk of the total rows in the list. To
 * keep acceleration while scrolling, we mark the nodes that are candidate for
 * deletion instead of deleting them right away, which would suddenly stop the
 * acceleration. We delete them once scrolling has finished.
 *
 * @param {Node} node Parent node where we want to append the children chunk.
 * @param {Number} from Starting position, i.e. first children index.
 * @return {void}
 */
VirtualList.prototype._renderChunk = function (node, from)
{
	var finalItem = from + this.cachedItemsLen;
	if (finalItem > this.totalRows)
		finalItem = this.totalRows;

	// Append all the new rows in a document fragment that we will later append to
	// the parent node
    var fragment = document.createDocumentFragment();

	for (var i = from; i < finalItem; i++)
    {
        var item = this.createRow(i);
        if (item.cached == false) {
            fragment.appendChild(item.item);
        }
		if (this.onRowCreated)
		{
			this.onRowCreated(i);
		}
	}

	node.appendChild(fragment);
};

VirtualList.createContainer = function (w, h)
{
	var c = document.createElement('div');
	c.style.width = w;
	c.style.height = h;
	c.style.position = 'relative';
	c.style.padding = 0;
	return c;
};

VirtualList.createScroller = function (h)
{
	var scroller = document.createElement('div');
	scroller.style.opacity = 0;
	scroller.style.position = 'absolute';
	scroller.style.top = 0;
	scroller.style.left = 0;
	scroller.style.width = '1px';
	scroller.style.height = h + 'px';
	return scroller;
};